#!/usr/bin/perl

use strict;
use warnings;
use Data::Dumper;

my $false = 0;
my $true = 1;
my $min_quarterword = 0;

my $right_brace=2;
my $spacer = 10;
my $letter=11;
my $other_char=12;

my $buffer = '';

my $cur_lang=-1;

sub set_cur_lang {  $cur_lang = shift; }

my $cur_chr;
my $cur_cmd;
my $buf_ptr = 0;

sub scan_left_brace {
  while (1) {
	$cur_chr = substr($buffer,$buf_ptr++,1);
	last if $cur_chr eq "{";
  }
};

sub get_x_token {
	$cur_chr = substr($buffer,$buf_ptr++,1);
	$cur_cmd = $other_char;
	$cur_cmd = $right_brace if $cur_chr eq '}';
	$cur_cmd = $spacer if $cur_chr eq ' ';
	$cur_cmd = $spacer if $cur_chr eq "\t";
	$cur_cmd = $letter if $cur_chr =~ /[a-zA-Z]/;
	$cur_chr = ord($cur_chr);
};

sub lc_code {
  my ($id) = shift;
  if ($id>96 && $id<=122) { return $id; }
  elsif ($id>64 && $id<=90) { return $id + 32;}
  else { return 0; }
}


#@* \[43] Initializing the hyphenation tables.
#The trie for \TeX's hyphenation algorithm is built from a sequence of
#patterns following a \.{\\patterns} specification. Such a specification
#is allowed only in \.{INITEX}, since the extra memory for auxiliary tables
#and for the initialization program itself would only clutter up the
#production version of \TeX\ with a lot of deadwood.
#
#The first step is to build a trie that is linked, instead of packed
#into sequential storage, so that insertions are readily made.
#After all patterns have been processed, \.{INITEX}
#compresses the linked trie by identifying common subtries. Finally the
#trie is packed into the efficient sequential form that the hyphenation
#algorithm actually uses.
#
#@<Declare subprocedures for |line_break|@>=
#@!init @<Declare procedures for preprocessing hyphenation patterns@>@;
#tini
#
#@ Before we discuss trie building in detail, let's consider the simpler
#problem of creating the |hyf_distance|, |hyf_num|, and |hyf_next| arrays.
#
#Suppose, for example, that \TeX\ reads the pattern `\.{ab2cde1}'. This is
#a pattern of length 5, with $n_0\ldots n_5=0\,0\,2\,0\,0\,1$ in the
#notation above. We want the corresponding |trie_op| code |v| to have
#|hyf_distance[v]=3|, |hyf_num[v]=2|, and |hyf_next[v]=@t$v^\prime$@>|,
#where the auxiliary |trie_op| code $v^\prime$ has
#|hyf_distance[@t$v^\prime$@>]=0|, |hyf_num[@t$v^\prime$@>]=1|, and
#|hyf_next[@t$v^\prime$@>]=min_quarterword|.
#
#\TeX\ computes an appropriate value |v| with the |new_trie_op| subroutine
#below, by setting
#$$\hbox{|@t$v^\prime$@>:=new_trie_op(0,1,min_quarterword)|,\qquad
#|v:=new_trie_op(3,2,@t$v^\prime$@>)|.}$$
#This subroutine looks up its three
#parameters in a special hash table, assigning a new value only if these
#three have not appeared before for the current language.
#
#The hash table is called |trie_op_hash|, and the number of entries it contains
#is |trie_op_ptr|.
#
#@<Glob...@>=

my %trie_op_hash=(); #  {trie op codes for quadruples}
my @trie_used=();    #  {largest opcode used so far for this language}
my @trie_op_lang=(); #  {language part of a hashed quadruple}
my @trie_op_val=();  #  {opcode corresponding to a hashed quadruple}
my $trie_op_ptr;   # {number of stored ops so far}

my $trie_size=200; #{space for hyphenation patterns; should be larger for  
                    # \.{INITEX} than it is in production versions of \TeX}
my $trie_op_size=75; #{space for ``opcodes'' in the hyphenation patterns}

my $max_languages = 24; # is normally 255

my $trie_not_ready;
my $trie_ptr;

my @trie_c=(); #  {characters to match}
my @trie_o=(); #  {operations to perform}
my @trie_l=(); #  {left subtrie links}
my @trie_r=(); #  {right subtrie links}
my @trie_hash=();  #  {used to identify equivalent subtries}
my @trie_ref=(); # {where linked trie families go into |trie|}

my @trie=(); # {|trie_link|, |trie_char|, |trie_op|}
my @hyf_distance=(); # {position |k-j| of $n_j$}
my @hyf_num=(); # {value of $n_j$}
my @hyf_next=(); # {continuation code}

my @trie_taken=(); #  {does a family start here?}
my @trie_min=(); #  {the first possible slot for each character}
my $trie_max=0; #{largest location used in |trie|}
my @op_start=(); # {offset for current language}

#@ It's tempting to remove the |overflow| stops in the following procedure;
#|new_trie_op| could return |min_quarterword| (thereby simply ignoring
#part of a hyphenation pattern) instead of aborting the job. However, that would
#lead to different hyphenation results on different installations of \TeX\
#using the same patterns. The |overflow| stops are necessary for portability
#of patterns.
#
#@<Declare procedures for preprocessing hyph...@>=
sub new_trie_op {
  my ($d,$n,$v) = @_;
  my $h; # {trial hash location}
  my $u; # {trial op code}
  my $l; # {pointer to stored data}
  $h=abs($n+313*$d+361*$v+1009*$cur_lang) % ($trie_op_size+$trie_op_size) - $trie_op_size;
  while (1) {
	$l=$trie_op_hash{$h};
	if ($l==0) { # {empty position found for a new op}
	  $u=$trie_used[$cur_lang];
	  $trie_op_ptr++;
	  $u++;
	  $trie_used[$cur_lang]=$u;
	  print OUT "\ntrie_used($cur_lang,$u);\n";
	  $hyf_distance[$trie_op_ptr]=$d;
	  print OUT "hyf_distance($trie_op_ptr,$d);\n";
	  $hyf_num[$trie_op_ptr]=$n;
	  print OUT "hyf_num($trie_op_ptr,$n);\n";
	  $hyf_next[$trie_op_ptr]=$v;
	  print OUT "hyf_next($trie_op_ptr,$v);\n";
	  $trie_op_lang[$trie_op_ptr]=$cur_lang;
	  print OUT "trie_op_lang($trie_op_ptr,$cur_lang);\n";
	  $trie_op_hash{$h}=$trie_op_ptr;
#	  print OUT "trie_op_hash($h,$trie_op_ptr);\n";
	  $trie_op_val[$trie_op_ptr]=$u;
	  print OUT "trie_op_val($trie_op_ptr,$u);\n\n";
	  return $u;
	}
	if (($hyf_distance[$l]==$d)&&($hyf_num[$l]==$n)&&($hyf_next[$l]==$v)
		&&($trie_op_lang[$l]==$cur_lang)) {
	  return $trie_op_val[$l];
	}
	if ($h>-$trie_op_size) {
	  $h--;
	} else {
	  $h=$trie_op_size;
	}
  }
}

#@ After |new_trie_op| has compressed the necessary opcode information,
#plenty of information is available to unscramble the data into the
#final form needed by our hyphenation algorithm.
#
#@<Sort \(t)the hyphenation op tables into proper order@>=
# see below
  
#@ Before we forget how to initialize the data structures that have been
#mentioned so far, let's write down the code that gets them started.
#
#@<Initialize table entries...@>=

for my $k (-$trie_op_size .. $trie_op_size) { $trie_op_hash{$k}=0; }
for my $k (0 .. 255) {  $trie_used[$k]=$min_quarterword; }
$trie_op_ptr=0;

$trie_not_ready=$true;
$trie_ptr=0;
$trie_c[0] = 0;
$trie_l[0] = 0;

my $img = 0;

#@ Now let's go back to the easier problem, of building the linked
#trie.  When \.{INITEX} has scanned the `\.{\\patterns}' control
#sequence, it calls on |new_patterns| to do the right thing.
#
#@<Declare procedures for preprocessing hyph...@>=
sub new_patterns { #{initializes the hyphenation pattern data}
  my ($langnum) = @_;
  my ($k,$l);       # {indices into |hc| and |hyf|; not always in |small_number| range}
  my $digit_sensed; # {should the next digit be treated as a letter?}
  my $v;            # {trie op code}
  my ($p,$q);       # {nodes of trie traversed during insertion}
  my $first_child;  # {is |p=trie_l[q]|?}
  my $c;            # {character being inserted}
  # the next two are really globals, but they are re-used as scratch
  # registers in this routine
  my @hyf = (); 
  my @hc=(0); # {word to be hyphenated}
  if ($trie_not_ready) {
	set_cur_lang($langnum);
	scan_left_brace(); # {a left brace must follow \.{\\patterns}}
	#@@<Enter all of the patterns into a linked trie, until coming to a right brace@>;
	$k=0;
	$hyf[0]=0;
	$digit_sensed=$false;
	while (1) {
      get_x_token();
	  if($cur_cmd==$letter||$cur_cmd==$other_char) {
	    #@@<Append a new letter or a hyphen level@>;
		if ($digit_sensed ||
			($cur_chr<ord("0")) ||
			($cur_chr>ord("9"))) {
		  if ($cur_chr==ord(".")) { # {edge-of-word delimiter}
			$cur_chr=0;
		  } else {
			$cur_chr=lc_code($cur_chr);
		  }
		  if ($k<63 ) {
			$k++;
			$hc[$k]=$cur_chr;
			$hyf[$k]=0;
			$digit_sensed=$false;
		  }
		} elsif ($k<63) {
		  $hyf[$k]=$cur_chr-ord("0");
		  $digit_sensed=$true;
		}
	  } elsif ($cur_cmd==$spacer||$cur_cmd==$right_brace) {
#		print OUT "cur_cmd_chr($cur_cmd,$cur_chr);\n";
		if ($#hc) {
		  print OUT "startimage($img);\n";
		  print OUT "hc(" . join(',',@hc) . ");\n"; #
		  print OUT "hyf(" . join(',',@hyf) . ");\n";
		  $img++;
		}
		if ($k>0) {
		  #@@<Insert a new pattern into the linked trie@>;
		  #@@<Compute the trie op code, |v|, and set |l:=0|@>;
		  if ($hc[1]==0)  {
			$hyf[0]=0;
		  }
		  if ($hc[$k]==0) {
			$hyf[$k]=0;
		  }
		  $l=$k;
		  $v=$min_quarterword;
		  while(1){
            if ($hyf[$l]!=0) { 
			  $v=new_trie_op($k-$l,$hyf[$l],$v);
			}
		    if ($l>0) {
			  $l--;
			} else {
			  last ;
			}
		  }
		  $q=0;
		  $hc[0]=$cur_lang;
		  while ($l<=$k) {
		    $c=$hc[$l];
			$l++;
			$p=$trie_l[$q];
			$first_child=$true;
		    while ($p>0 && $c>$trie_c[$p]) {
		      $q=$p;
			  $p=$trie_r[$q];
			  $first_child=$false;
			}
		    if ($p==0 || $c<$trie_c[$p]) {
			  #@@<Insert a new trie node between |q| and |p|, and make |p| point to it@>;
			  $trie_ptr++;
			  $trie_r[$trie_ptr]=$p;
			  print OUT "trie_r($trie_ptr,$p);\n";
			  $p=$trie_ptr;
			  $trie_l[$p]=0;
			  print OUT "trie_l($p,0);\n";
			  if ($first_child) {
				$trie_l[$q]=$p;
				print OUT "trie_l($q,$p);\n";
			  } else {
				$trie_r[$q]=$p;
				print OUT "trie_r($q,$p);\n";
			  }
			  $trie_c[$p]=$c;
			  print OUT "trie_c($p,$c); % ". chr($c). "\n";
			  $trie_o[$p]=$min_quarterword;
			  print OUT "trie_o($p,$min_quarterword)\n";
			}
		    $q=$p; # {now node |q| represents $p_1\ldots p_{l-1}$}
		  }
		  $trie_o[$q]=$v;
		  print OUT "trie_o($q,$v);\n";
		}
		if ($#hc) {
		  print OUT "stopimage;\n\n";
		}
		# easier for MP display ...
		@hyf = (); 
		@hc=(0); # {word to be hyphenated}
		# 
		if ($cur_cmd==$right_brace) {
		  last;
		}
	    $k=0;
		$hyf[0]=0;
		$digit_sensed=$false;
	  }
	}
  }
}

sub compress_trie{ #;
  my ($p) = @_;
  if ($p==0) {
	return 0;
  } else {
	my $val = compress_trie($trie_l[$p]);
	$trie_l[$p]= $val;
    print OUT "startimage($img);\n"; $img++;
	print OUT "trie_l($p, $val);\n";
    print OUT "stopimage;\n";
	$val = compress_trie($trie_r[$p]);
    $trie_r[$p]=$val;
    print OUT "startimage($img);\n"; $img++;
	print OUT "trie_r($p, $val);\n";
    print OUT "stopimage;\n";
    return trie_node($p);
  }
}

sub trie_root {
  my $val = shift;
  if (defined $val) {
	$trie_l[0] = $val;
	print OUT "trie_l(0,$val);\n";
  }
  $trie_l[0];
} # {root of the linked trie}


sub trie_node { #; {converts to a canonical form}
  my ($p) = @_;
  my $h; #{trial hash location}
  my $q; #{trial trie node}
  $h=abs($trie_c[$p]+1009*$trie_o[$p]+2718*$trie_l[$p]+3142*$trie_r[$p]) %  $trie_size;
  while(1){
	$q=$trie_hash[$h];
	if ($q==0)  {
	  $trie_hash[$h]=$p;
      print OUT "startimage($img);\n"; $img++;
	  print OUT "trie_hash($h,$p);\n";
      print OUT "stopimage;\n";
	  return $p;
	}
	if (($trie_c[$q]==$trie_c[$p])&&($trie_o[$q]==$trie_o[$p])&&
		($trie_l[$q]==$trie_l[$p])&&($trie_r[$q]==$trie_r[$p])) {
	  return $q;
	}
	if ($h>0) {$h--; } else { $h=$trie_size; }
  }
}

sub trie_link { # {``downward'' link in a trie}
  my ($a,$val) = @_;
  if (defined $val) {
	$trie[$a]->{'rh'} = $val;
	print OUT "trie_rh($a,$val);\n";
  }
  $trie[$a]->{'rh'};
}
sub trie_char { # {character matched at this trie location}
  my ($a,$val) = @_;
  if (defined $val) {
	$trie[$a]->{'b1'} = $val;
	print OUT "trie_one($a,$val);\n";
  }
  $trie[$a]->{'b1'};
}
sub trie_op { # {program for hyphenation at this trie location}
  my ($a,$val) = @_;
  if (defined $val) {
	$trie[$a]->{'b0'} = $val;
	print OUT "trie_zero($a,$val);\n";
  }
  $trie[$a]->{'b0'};
}


sub first_fit { # {packs a family into |trie|}
  my ($p) = @_;
  my $h; #{candidate for |trie_ref[p]|}
  my $z; #{runs through holes}
  my $q; #{runs through the family starting at |p|}
  my $c; #{smallest character in the family}
  my ($l,$r); #{left and right neighbors}
  my $ll; #{upper limit of |trie_min| updating}
  $c=$trie_c[$p];
  $z=$trie_min[$c]; # {get the first conceivably good hole}
 NEW:
  while(1) {
	$h=$z-$c;
	#@@<Ensure that |trie_max>=h+256|@>;
	if ($trie_max<$h+256) {
	  while($trie_max<=$h+256) { 
		$trie_max++;
		$trie_taken[$trie_max]=$false;
		print OUT "startimage($img);\n"; $img++;
		print OUT "trie_taken($trie_max,$false);\n";
	    trie_link($trie_max,$trie_max+1);
		trie_back($trie_max,$trie_max-1);
        print OUT "stopimage;\n";
	  };
	}
	if ($trie_taken[$h]) { 
	  $z=trie_link($z); # {move to the next hole}
	  next NEW;
	}
	#@@<If all characters of the family fit relative to |h|, then..@>
	$q=$trie_r[$p];
	while ($q>0) {
	  if (trie_link($h+$trie_c[$q])==0) { 
		$z=trie_link($z); # {move to the next hole}
		next NEW;
	  }
	  $q=$trie_r[$q];
	}
	last;
  }
  #@@<Pack the family into |trie| relative to |h|@>;
  $trie_taken[$h]=$true;
  print OUT "startimage($img);\n"; $img++;
  print OUT "trie_taken($h,$true);\n";
  $trie_ref[$p]=$h;
  print OUT "trie_ref($p,$h);\n";
  print OUT "stopimage;\n";
  $q=$p;
  while($q!=0) {
    $z=$h+$trie_c[$q];
	$l=trie_back($z);
	$r=trie_link($z);
	print OUT "startimage($img);\n"; $img++;
    trie_back($r,$l);
	trie_link($l,$r);
	trie_link($z,0);
    if ($l<256) {
	  if ($z<256) {
		$ll=$z;
	  } else {
		$ll=256;
	  }
      do {
		$trie_min[$l]=$r;
		print OUT "trie_min($l,$r);\n";
		$l++;
	  } until $l==$ll;
	}
    print OUT "stopimage;\n";
    $q=$trie_r[$q];
  };
}

sub trie_back { #{backward links in |trie| holes}
  my ($a,$val ) = @_;
  if (defined $val) {
	$trie[$a]->{'lh'} = $val;
	print OUT "trie_lh($a,$val);\n";
  }
  $trie[$a]->{'lh'}; 
}

sub trie_pack { # {pack subtries of a family}
  my ($p) = @_;
  my $q; # {a local variable that need not be saved on recursive calls}
  while($p!=0) {
    $q=$trie_l[$p];
    if (($q>0)&&($trie_ref[$q]==0)) {
	  first_fit($q);
	  trie_pack($q);
	}
    $p=$trie_r[$p];
  };
}

sub trie_fix { # {moves |p| and its siblings into |trie|}
  my ($p) = @_;
  my $q; # {a local variable that need not be saved on recursive calls}
  my $c; # {another one that need not be saved}
  my $z; # {|trie| reference; this local variable must be saved}
  $z=$trie_ref[$p];
  do {
    $q=$trie_l[$p];
	$c=$trie_c[$p];
	print OUT "startimage($img);\n"; $img++;
	trie_link($z+$c,$trie_ref[$q]);
	trie_char($z+$c,$c);
	trie_op($z+$c,$trie_o[$p]);
	print OUT "stopimage;\n";
    if ($q>0) {
	  trie_fix($q); 
	}
    $p=$trie_r[$p];
  } until $p==0;
}

sub init_trie {
  my $p; # {pointer for initialization}
  my ($j,$k,$t); # {all-purpose registers for initialization}
  my ($r,$s); # {used to clean up the packed |trie|}
  my $h; # {template used to zero out |trie|'s holes}
  #@@<Get ready to compress the trie@>;
  #@@<Sort \(t)the hyphenation...@>;
  $op_start[0]=-$min_quarterword;
  print OUT "startimage($img);\n";  $img++;
  print OUT "init_trie;\n";
  print OUT "stopimage;\n";
  print OUT "startimage($img);\n"; $img++;
  print OUT "op_start(0,$min_quarterword);\n";
  for $j (1 .. $max_languages) {
	$op_start[$j]=$op_start[$j-1]+$trie_used[$j-1];
	print OUT "op_start($j,". ($op_start[$j-1]+$trie_used[$j-1]) .");\n";
  }
  print OUT "stopimage;\n";
  print OUT "startimage($img);\n"; $img++;
  for $j (1 .. $trie_op_ptr) { 
	$trie_op_hash{$j}=$op_start[$trie_op_lang[$j]]+$trie_op_val[$j];   # {destination}
	print OUT "trie_op_hash($j," . ($op_start[$trie_op_lang[$j]]+$trie_op_val[$j]) . ");\n";
  }
  print OUT "stopimage;\n";
  for $j (1 .. $trie_op_ptr) {
	while ($trie_op_hash{$j}>$j) {
	  # in this case, pattern files were offered to the machinery
	  # in a non sequential manner (patterns for a low-numbered language follow 
      # after those of a higher numbered language).
	  print OUT "startimage($img);\n"; $img++;
	  $k=$trie_op_hash{$j};
	  $t=$hyf_distance[$k];
	  $hyf_distance[$k]=$hyf_distance[$j];
	  print OUT "hyf_distance($k,".$hyf_distance[$j]. ");\n";
	  $hyf_distance[$j]=$t;
	  print OUT "hyf_distance($j,$t);\n";
	  $t=$hyf_num[$k];
	  $hyf_num[$k]=$hyf_num[$j];
	  print OUT "hyf_num($k,".$hyf_num[$j]. ");\n";
	  $hyf_num[$j]=$t;
	  print OUT "hyf_num($j,$t);\n";
	  $t=$hyf_next[$k];
	  $hyf_next[$k]=$hyf_next[$j];
	  print OUT "hyf_next($k,".$hyf_next[$j]. ");\n";
	  $hyf_next[$j]=$t;
	  print OUT "hyf_next($j,$t);\n";
	  $trie_op_hash{$j}=$trie_op_hash{$k};
	  print OUT "trie_op_hash($j,".$trie_op_hash{$k}. ");\n";
	  $trie_op_hash{$k}=$k;
	  print OUT "trie_op_hash($k,$k);\n";
	  print OUT "stopimage;\n";
	}
  }
  print OUT "startimage($img);\n"; $img++;
  for $p (0 .. $trie_size) { # cheating, for display
	$trie_hash[$p]=0;
	print OUT "trie_hash($p,0);\n" unless $p == $trie_size; # cheating, for display
  }
  print OUT "stopimage;\n";
  trie_root(compress_trie(trie_root())); # {identify equivalent subtries}
  print OUT "startimage($img);\n"; $img++;
  for $p (0 .. $trie_ptr) {
	$trie_ref[$p]=0;
	print OUT "trie_ref($p,0);\n";
  }
  print OUT "stopimage;\n";
  print OUT "startimage($img);\n"; $img++;
  for $p (0 .. 255) {
	$trie_min[$p]=$p+1;
	print OUT "trie_min($p,".($p+1).");\n";
  }
  print OUT "stopimage;\n";
  print OUT "startimage($img);\n"; $img++;
  trie_link(0,1);
  print OUT "stopimage;\n";
  $trie_max=0;
  if (trie_root()!=0) {
	first_fit(trie_root());
	trie_pack(trie_root());
  }
  #@@<Move the data into |trie|@>;
  $h->{'rh'}=0;
  $h->{'b0'}=$min_quarterword;
  $h->{'b1'}=$min_quarterword; # {|trie_link:=0|,
    #|trie_op:=min_quarterword|, |trie_char:=qi(0)|}
  if (trie_root()==0) { #  {no patterns were given}
	print OUT "startimage($img);\n"; $img++;
	for $r (0 .. 256) {
	  $trie[$r]={%{$h}};
	  print OUT "trie_one($r,". $h->{b1}.");\n";
	  print OUT "trie_zero($r,".$h->{b0}.");\n";
	  print OUT "trie_rh($r," . $h->{rh}.");\n";
	}
	$trie_max=256;
	print OUT "stopimage;\n";
  } else {
	trie_fix(trie_root()); # {this fixes the non-holes in |trie|}
	$r=0; # {now we will zero out all the holes}
	do {
	  $s=trie_link($r);
	  $trie[$r]={%{$h}};
	  print OUT "startimage($img);\n"; $img++;
	  print OUT "trie_one($r,". $h->{b1}.");\n";
	  print OUT "trie_zero($r,".$h->{b0}.");\n";
	  print OUT "trie_rh($r," . $h->{rh}.");\n";
      print OUT "stopimage;\n";
	  $r=$s;
	} until $r>$trie_max;
  }
  print OUT "startimage($img);\n"; $img++;
  trie_char(0,ord("?")); #  {make |trie_char(c)!=c| for all |c|}
  print OUT "stopimage;\n";
  $trie_not_ready=$false;
}


sub read_patterns {
  my ($filename,$num) = @_;
  open(IN,"<$filename");
  while (my $line = <IN>) {
	chomp $line;
	$line =~ s/%.*$//;
	$line =~ s/\s*$/ /;
	$buffer .= $line;
  }
  close IN;
  while (length($buffer)) {
	$buffer =~ s/^\s//;
	if ($buffer =~ s/^\\patterns//) {
	  new_patterns($num);
	  $buffer = substr($buffer,$buf_ptr);
	  $buf_ptr = 0;
	} elsif ($buffer =~ s/^\\hyphenation//) {
	  $buffer =~ s/\{[^}]*\}//;
	}
  }
}

sub my_hyf {
  my ($word,$cur_lang,$l_hyf,$r_hyf) = @_;
  $cur_lang ||= 0;
  my @chars = map {ord($_)} split('',$word);
  my $hn = length($word);
  my @hc = (0,@chars,0,256);
  my @hyf;
  foreach my $i (0.. ($hn+1)) {
	$hyf[$i] = 0;
  }
  # {insert delimiters}
  for my $j (0 .. ($hn-$r_hyf+1)) {
    my $z=trie_link($cur_lang+1)+$hc[$j];
	my $l=$j;
    while ($hc[$l]==trie_char($z)) {
      if (trie_op($z)!=$min_quarterword) {
		#@@<Store \(m)maximum values in the |hyf| table@>;
		my $v=trie_op($z);
		while($v!=$min_quarterword) {
          $v=$v+$op_start[$cur_lang];
		  my $i=$l-$hyf_distance[$v];
		  if ($hyf_num[$v]>$hyf[$i]) {
			$hyf[$i]=$hyf_num[$v]; 
		  }
		  $v=$hyf_next[$v];
		};
	  }
	  $l++;
	  $z=trie_link($z)+$hc[$l];
	}
  }
  my $char = 0;
  my $ret = '';
  while ($char < $hn+1) {
	if ($hc[$char] > 0) {
	  $ret .= chr($hc[$char]);
	  if (($char >= $l_hyf) && $hyf[$char] % 2 && $char <= ($hn-$r_hyf)) {
		$ret .= '-';
	  }
	}
	$char++;
  }
  return $ret;
}


sub hyphenate {
  my ($par) = @_;
  my @words = split (/ /,$par);
  my @hy;
  while (my $word = shift @words) {
	push @hy, my_hyf($word,0,2,3); # $lang,$l_hyf,$r_hyf
  }
  return join (' ', @hy) . "\n";
  
}

open (OUT,">pat.mp");
#read_patterns("hyphen.tex",0);
read_patterns("langa.tex",0);
read_patterns("langabc.tex",2);
read_patterns("langb.tex",1);
init_trie();
$cur_lang = 0;
if (@ARGV)  {
  open (IN,"<" . shift @ARGV);
  while (<IN>) {
	chomp;
    print hyphenate($_);
  }
  close IN;
}
